@@ -1,20 +1,74 @@
 from zope import interface
 
-from repoze.formapi import interfaces
 from repoze.formapi import marshalling
+from repoze.formapi.error import Errors
 
 import types
 import re
 
+def get_instances_of(type, *bases):
+    for base in bases:
+        for value in base.__dict__.values():
+            if isinstance(value, type):
+                yield value
+        for value in get_instances_of(type, *base.__bases__):
+            yield value            
+
+class Validator(object):
+    """Wrapper for validators.
+
+    This calls a validator object (usually a method) and collects all it's
+    errors. It also sets a flag that the form library can use to know that it
+    is a validator."""
+
+    def __init__(self, func, *fields):
+        self.func = func
+        # Fields can be empty, in that case we want to have and empty path
+        if not fields:
+            self.fieldpaths = ((),)
+        else:
+            self.fieldpaths = [f.split('.') for f in fields]
+
+    def __call__(self, form):
+        for error in self.func(form):
+            for fieldpath in self.fieldpaths:
+                yield (fieldpath, error)
+
+class Action(object):
+    def __init__(self, action, name=None, submitted=False):
+        self.action = action
+        self.name = name
+        self.submitted = submitted
+
+    @property
+    def __call__(self):
+        return self.action
+
+    def __nonzero__(self):
+        return self.submitted
+
+    def __repr__(self):
+        return '<%s name="%s" submitted="%s">' % (
+            type(self).__name__, self.name or "", str(bool(self.submitted)))
+    
+class metaclass(type):
+    def __init__(kls, name, bases, dict):
+        kls.validators = tuple(get_instances_of(Validator, kls))
+        kls.actions = tuple(get_instances_of(Action, kls))
+
 class Form(object):
     """Base form class. Optionally pass a dictionary as ``data`` and a
     WebOb-like request object as ``request``."""
 
-    interface.implements(interfaces.IForm)
+    __metaclass__ = metaclass
     
     fields = {}
+    status = None
     
     def __init__(self, data=None, context=None, request=None, prefix=None):
+        self.context = context
+        self.request = request
+
         if context is not None:
             if data is not None:
                 raise ValueError(
@@ -24,7 +78,7 @@ class Form(object):
             data = Proxy(context)
 
         self.data = Data(data)
-
+        
         # find action parameters
         if prefix is not None:
             re_prefix = re.compile(r'^%s[._-](?P<name>.*)' % prefix)
@@ -39,12 +93,11 @@ class Form(object):
 
         # initialize form actions
         actions = self.actions = []
-        for action in type(self).__dict__.values():
-            marker = object()
-            name = getattr(action, '__action__', marker)
-            if name is not marker:
-                actions.append(Action(action, name, name in action_params))                
-
+        for action in type(self).actions:
+            name = action.name
+            actions.append(
+                Action(action.__call__, name, name in action_params))
+                
         # conditionally apply request parameters if:
         # 1. no prefix has been set
         # 2. there is a submitted action
@@ -56,11 +109,13 @@ class Form(object):
             params = request.params.items()
         else:
             params = ()
-        
+
         # marshall request parameters
         data, errors = marshalling.marshall(params, self.fields)
-        
-        self.data.update(data)
+
+        if len(params):
+            self.data.update(data)
+
         self.errors = errors
         self.prefix = prefix
 
@@ -69,23 +124,20 @@ class Form(object):
 
         for action in self.actions:
             if action:
-                return action(self, self.data)
-        
+                status = self.status = action(self, self.data)
+                return status
+            
     def validate(self):
         """Validates the request against the form fields. Returns
         ``True`` if all fields validate, else ``False``."""
+        for validator in self.validators:
+            for field_path, validation_error in validator(self):
+                errors = self.errors
+                for field in field_path:
+                    errors = errors[field]
+                errors += validation_error
 
-        if self.errors:
-            return False
-
-        # execute custom form validators
-        for validator in type(self).__dict__.values():
-            if getattr(validator, '__validator__', False):
-                for error in validator(self):
-                    path = tuple(error.field.split('.'))
-                    self.errors[path] = error.msg
-
-        return not self.errors
+        return not bool(self.errors)
 
 class ValidationError(Exception):
     """Represents a field validation error."""
@@ -102,21 +154,21 @@ class ValidationError(Exception):
 
     def __str__(self):
         return str(unicode(self))
-    
+
     def __unicode__(self):
         return self.msg
-    
+
 class Data(list):
     """Form data object with dictionary-like interface. If initialized
     with a ``data`` object, this will be used to provide default
     values, if not set in the ``request``. Updates to the object are
     transient until the ``save`` method is invoked."""
-    
+
     def __init__(self, data):
         if data is not None:
             self.append(data)
         self.append({})
-        
+
     def __getitem__(self, name):
         for data in reversed(self):
             try:
@@ -126,7 +178,7 @@ class Data(list):
 
             if value is not None:
                 return value
-            
+    
     def __setitem__(self, name, value):
         self.tail[name] = value
 
@@ -137,13 +189,13 @@ class Data(list):
     @property
     def head(self):
         return list.__getitem__(self, 0)
-        
+
     def update(self, data):
         """Updates the dictionary by appending ``data`` to the list at
         the position just before the current dictionary."""
 
         self.insert(-1, data)
-        
+
     def save(self):
         """Flattens the dictionary, saving changes to the data object."""
 
@@ -152,24 +204,6 @@ class Data(list):
                 self.head[name] = value
         self.append({})
 
-class Action(object):
-    """Callable form action."""
-    
-    def __init__(self, action, name, submitted):
-        self.name = name
-        self.submitted = submitted
-        self.__call__ = action
-
-    def __nonzero__(self):
-        return self.submitted
-
-    def __repr__(self):
-        return '<%s name="%s" submitted="%s">' % (
-            type(self).__name__, self.name or "", str(bool(self.submitted)))
-
-    def __call__(self, *args):
-        return self.__call__(*args)
-        
 class Proxy(object):
     """Proxy object; reads and writes to attributes are forwarded to
     the provided object. Descriptors are supported: they must read and
@@ -243,7 +277,7 @@ class Proxy(object):
         else:
             return getattr(
                 object.__getattribute__(self, '_context'), name)
-        
+
     def __setattr__(self, name, value):
         prop = object.__getattribute__(type(self), '__dict__').get(name)
         if prop is not None:
@@ -259,32 +293,19 @@ class Proxy(object):
 
     __getitem__ = __getattribute__
     __setitem__ = __setattr__
-
-def validator(*fields):
-    def decorator(validator):
-        def func(self):
-            result = validator(self)
-            if result is not None:
-                if not isinstance(result, (
-                    types.TupleType, types.ListType, types.GeneratorType)):
-                    result = (result,)
-                for field in fields:
-                    for msg in result:
-                        yield ValidationError(field, msg)
-        func.__validator__ = True
-        return func
-    return decorator
-
+    
 def action(name):
     if isinstance(name, types.FunctionType):
-        def func(self, data):
-            return name(self, data)
-        func.__action__ = None
-        return func
+        return Action(name)
     else:
         def decorator(action):
-            def func(self, data):
-                return action(self, data)
-            func.__action__ = name
-            return func
+            return Action(action, name)
         return decorator
+
+def validator(*args):
+    # If the first (and only) argument is a callable process it
+    if len(args) == 1 and callable(args[0]):
+        return Validator(args[0])
+    # Treat the args as field names and prepare a wrapper
+    else:
+        return lambda func: Validator(func, *args)
